; -----------------------------------------------------------------------------
; RetroForth for Linux       | This port was originally written by:
; http://www.retroforth.org  | * Tom Novelli *
; -----------------------------------------------------------------------------
macro section a { }
include "macros.fasm"

format elf executable
entry main

main:
	call init_retroforth
	call interpret
;---------------------------------------------------------------------
code 'emit', emit	
	push ebx,ecx,edx
	mov ebx, 1	;stdout
	mov edx, ebx	;count
	lea ecx, [esp-4]
	mov [ecx], eax
	mov eax, 4	;sys_write
	int 80h
	pop ebx,ecx,edx
	drop
	ret
code 'key', key
	dup
	push ebx,ecx,edx
	sub ebx,ebx	;0 = stdin
	mov edx, 1	;count
	lea ecx, [esp-4]
	mov eax, 3	;sys_read
	int 80h
	mov eax, [ecx]
	pop ebx,ecx,edx
	ret
;---------------------------------------------------------------------
code 'bye', linux_bye	
	sub ebx,ebx	;Return exit status 0
	mov eax,ebx
	inc eax		;1 = sys_exit
	int 80h

; Linux syscall interface
; Description:
; syscal ( args argcount syscall-number -- result )
; args: as few as 0 or as many as 6 arguments
; argcount: the number of arguments
; syscall-number: a number from /usr/include/asm/unistd.h
; This calling convention was borrowed from Marcel Hendrix's port of eforth
; for Linux.

code 'syscall', linux_syscall

	; syscalls can take a variable number of arguments, from 0 to 6.
	; The args go in ebx, ecx, edx, esi, edi, ebp, in that order.
	; (this is from the Linux Assembly HOWTO)
	
	; At entry, the Forth stack holds args, argcount, syscall-number
	push eax
	
	drop	; remove syscall-number from the Forth stack
	cmp eax, 0	; no args
	je lsc_0arg
	cmp eax, 1
	je lsc_1arg
	cmp eax, 2
	je lsc_2arg
	cmp eax, 3
	je lsc_3arg
	cmp eax, 4
	je lsc_4arg
	cmp eax, 5
	je lsc_5arg
	cmp eax, 6
	je lsc_6arg
	
	; fallthrough.  argcount is not in the interval [0, 6].
	; The Forth stack is restored to the state it had before syscall.
	sub esi, 4
	mov [esi], eax
	pop eax
	next
	
	lsc_0arg:
		pop eax	; the syscall number is refetched.  argcount is gone
		int 80h
		next
		
	lsc_1arg:
		pop eax
		push ebx
		
		; From now on, we do lots of register exchanges,
		; to keep the arguments in their places.
		xchg ebx, eax
		drop
		xchg ebx, eax
		int 80h
		pop ebx
		next
		
	lsc_2arg:
		pop eax
		push ebx, ecx
		xchg eax, ecx
		drop
		mov ebx, eax
		drop
		xchg eax, ecx
		int 80h
		pop ebx, ecx
		next
		
	lsc_3arg:
		pop eax
		push ebx, ecx, edx
		xchg eax, edx
		drop
		upop ebx
		upop ecx
		xchg eax, edx
		int 80h
		pop ebx, ecx, edx
		next
		
		; 4, 5, and 6 argument syscalls need to preserve esi.
		; While using mmap and ipc, I noticed that the data
		; stack was getting corrupted.  Fixed.
	lsc_4arg:
		pop eax
		push ebx, ecx, edx
		xchg eax, edx
		drop
		upop ebx
		upop ecx
		xchg eax, edx
		push eax
		drop

		; argument 4 in eax, syscall number at [esp]
xchg esi, [esp]
xchg esi, eax
				; Data stack pointer in [esp], syscall number in eax,
				; argument 4 in esi.

		int 80h
		pop ebx, ecx, edx, esi
		next
		
	lsc_5arg:
		pop eax
		push ebx, ecx, edx, edi
		push eax
		drop
		upop ebx
		upop ecx
		upop edx
		upop edi
		xchg eax, esi
		xchg edi, esi

		; arguments 1 through 5 are in place.  [esp] holds
		; the syscall number.  The data stack pointer is in eax
		xchg eax, [esp]
		int 80h
		pop ebx, ecx, edx, edi, esi
		next
		
	lsc_6arg:
		pop eax
		push ebx, ecx, edx, edi, ebp
		push eax
		drop
		upop ebx
		upop ecx
		upop edx
		upop edi
		upop ebp
		xchg ebp, eax
		xchg edi, eax
		xchg esi, eax
		xchg eax, [esp]
		int 80h
		pop ebx, ecx, edx, edi, ebp, esi
		next
; syscall interface ends here


;---------------------------------------------------------------------
include "retroforth.asm"
